{% extends "!layout.html" %}

{% block rootrellink %}
<li><a href="https://www.breezy-vcs.org/">
	<!--TODO(jelmer): Add a brz icon<img src="{{ pathto("_static/brz icon 16.png", 1) }}" /> -->
	Home</a>&nbsp;|&nbsp;</li>
<a href="https://www.breezy-vcs.org/doc/en/">Documentation</a>&nbsp;|&nbsp;</li>
{{ super() }}
{% endblock %}
